/*	$NetBSD$	*/

/*-
 * Copyright (c) 2020 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	_CRYPTO_AES_ARCH_ARM_AES_NEON_INTERNAL_H
#define	_CRYPTO_AES_ARCH_ARM_AES_NEON_INTERNAL_H

#include <sys/types.h>

#if 1
#include <arm_neon.h>
#ifndef __aarch64__
static inline uint8x16_t
vqtbl1q_u8(uint8x16_t tab, uint8x16_t idx)
{
	uint64x2_t tab64 = (uint64x2_t)tab;
	uint64x2_t idx64 = (uint64x2_t)idx;
	uint8x8_t tablo = (uint8x8_t)tab64[0];
	uint8x8_t tabhi = (uint8x8_t)tab64[1];
	uint8x8_t idxlo = (uint8x8_t)idx64[0];
	uint8x8_t idxhi = (uint8x8_t)idx64[1];
	uint8x8x2_t tab8x8x2 = { { tablo, tabhi } };
	uint8x8_t outlo, outhi;
	uint64x2_t out64;

	outlo = vtbl2_u8(tab8x8x2, idxlo);
	outhi = vtbl2_u8(tab8x8x2, idxhi);
	out64 = (uint64x2_t) { (uint64x1_t)outlo, (uint64x1_t)outhi };

	return (uint8x16_t)out64;
}
#endif
#else
#include "arm_neon.h"
#endif

#include <crypto/aes/aes.h>

#include "aes_neon_subr.h"

uint8x16_t aes_neon_enc1(const struct aesenc *, uint8x16_t, unsigned);
uint8x16_t aes_neon_dec1(const struct aesdec *, uint8x16_t, unsigned);

#endif	/* _CRYPTO_AES_ARCH_ARM_AES_NEON_INTERNAL_H */
