/*	$NetBSD$	*/

/*-
 * Copyright (c) 2020 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	_SYS_CRYPTO_AES_ARCH_ARM_ARM_NEON_H
#define	_SYS_CRYPTO_AES_ARCH_ARM_ARM_NEON_H

#if defined(__GNUC__) && !defined(__clang__)

#define	_INTRINSATTR							      \
	__extension__							      \
	__attribute__((__always_inline__, __gnu_inline__, __artificial__))

#ifdef __aarch64__
typedef __Int32x4_t int32x4_t;
typedef __Int64x2_t int64x2_t;
typedef __Int8x16_t int8x16_t;
typedef __Uint32x4_t uint32x4_t;
typedef __Uint64x2_t uint64x2_t;
typedef __Uint8x16_t uint8x16_t;
#else
typedef __simd128_int32_t int32x4_t;
typedef __simd128_int64_t int64x2_t;
typedef __simd128_int8_t int8x16_t;
typedef __simd128_uint32_t uint32x4_t;
typedef __simd128_uint64_t uint64x2_t;
typedef __simd128_uint8_t uint8x16_t;

typedef __simd64_int8_t int8x8_t;
typedef __simd64_uint8_t uint8x8_t;
typedef __builtin_neon_udi uint64x1_t;
typedef struct { uint8x8_t val[2]; } uint8x8x2_t;
#endif

#if defined(__AARCH64EB__) || defined(__ARM_BIG_ENDIAN)
#define	__neon_lane_index(__v, __i)	(__arraycount(__v) - 1 - __i)
#else
#define	__neon_lane_index(__v, __i)	__i
#endif

#elif defined(__clang__)

#error Teach me how to neon in clang!

#else

#error Teach me how to neon in your compile!

#endif

_INTRINSATTR
static __inline uint32x4_t
vcltq_s32(int32x4_t __v0, int32x4_t __v1)
{
	return (uint32x4_t)(__v0 < __v1);
}

_INTRINSATTR
static __inline int32x4_t
vdupq_n_s32(int32_t __x)
{
	return (int32x4_t) { __x, __x, __x, __x };
}

_INTRINSATTR
static __inline uint32x4_t
vdupq_n_u32(uint32_t __x)
{
	return (uint32x4_t) { __x, __x, __x, __x };
}

_INTRINSATTR
static __inline uint8x16_t
vdupq_n_u8(uint8_t __x)
{
	return (uint8x16_t) {
		__x, __x, __x, __x, __x, __x, __x, __x,
		__x, __x, __x, __x, __x, __x, __x, __x,
	};
}

_INTRINSATTR
static __inline uint32x4_t
vextq_u32(uint32x4_t __lo, uint32x4_t __hi, uint8_t __i)
{
#if defined(__AARCH64EB__) || defined(__ARM_BIG_ENDIAN)
	return __builtin_shuffle(__hi, __lo,
	    (uint32x4_t) { 4 - __i, 5 - __i, 6 - __i, 7 - __i });
#else
	return __builtin_shuffle(__lo, __hi,
	    (uint32x4_t) { __i + 0, __i + 1, __i + 2, __i + 3 });
#endif
}

_INTRINSATTR
static __inline uint8x16_t
vextq_u8(uint8x16_t __lo, uint8x16_t __hi, uint8_t __i)
{
#if defined(__AARCH64EB__) || defined(__ARM_BIG_ENDIAN)
	return __builtin_shuffle(__hi, __lo,
	    (uint8x16_t) {
		16 - __i, 17 - __i, 18 - __i, 19 - __i,
		20 - __i, 21 - __i, 22 - __i, 23 - __i,
		24 - __i, 25 - __i, 26 - __i, 27 - __i,
		28 - __i, 29 - __i, 30 - __i, 31 - __i,
	});
#else
	return __builtin_shuffle(__lo, __hi,
	    (uint8x16_t) {
		__i +  0, __i +  1, __i +  2, __i +  3,
		__i +  4, __i +  5, __i +  6, __i +  7,
		__i +  8, __i +  9, __i + 10, __i + 11,
		__i + 12, __i + 13, __i + 14, __i + 15,
	});
#endif
}

_INTRINSATTR
static __inline uint32_t
vgetq_lane_u32(uint32x4_t __v, uint8_t __i)
{
#ifdef __aarch64__
	return __v[__i];
#else
	return (uint32_t)__builtin_neon_vget_laneuv4si((int32x4_t)__v, __i);
#endif
}

_INTRINSATTR
static __inline uint8x16_t
vld1q_u8(const uint8_t *__p8)
{
#ifdef __aarch64__
	const __builtin_aarch64_simd_qi *__p =
	    (const __builtin_aarch64_simd_qi *)__p8;

	return (uint8x16_t)__builtin_aarch64_ld1v16qi(__p);
#else
	const __builtin_neon_qi *__p = (const __builtin_neon_qi *)__p8;

	return (uint8x16_t)__builtin_neon_vld1v16qi(__p);
#endif
}

_INTRINSATTR
static __inline uint8x16_t
vqtbl1q_u8(uint8x16_t __tab, uint8x16_t __idx)
{
#ifdef __aarch64__
	uint8x16_t __res;
	__asm__("tbl %0.16b, {%1.16b}, %2.16b"
	    : "=w"(__res) : "w"(__tab), "w"(__idx));
	return __res;
#else
	/*
	 * No native ARMv7 NEON instruction for this, so do it via two
	 * half-width TBLs instead (vtbl2_u8 equivalent).
	 */
	uint64x2_t __tab64 = (uint64x2_t)__tab;
	uint8x8_t __tablo = (uint8x8_t)__tab64[0];
	uint8x8_t __tabhi = (uint8x8_t)__tab64[1];
	uint8x8x2_t __tab8x8x2 = { { __tablo, __tabhi } };
	union {
		uint8x8x2_t __u8x8x2;
		__builtin_neon_ti __ti;
	} __u = { __tab8x8x2 };
	uint64x2_t __idx64, __out64;
	int8x8_t __idxlo, __idxhi, __outlo, __outhi;

	__idx64 = (uint64x2_t)__idx;
	__idxlo = (int8x8_t)__idx64[0];
	__idxhi = (int8x8_t)__idx64[1];
	__outlo = (int8x8_t)__builtin_neon_vtbl2v8qi(__u.__ti, __idxlo);
	__outhi = (int8x8_t)__builtin_neon_vtbl2v8qi(__u.__ti, __idxhi);
	__out64 = (uint64x2_t) { (uint64x1_t)__outlo, (uint64x1_t)__outhi };

	return (uint8x16_t)__out64;
#endif
}

_INTRINSATTR
static __inline int32x4_t
vreinterpretq_s32_u8(uint8x16_t __v)
{
	return (int32x4_t)__v;
}

_INTRINSATTR
static __inline uint32x4_t
vreinterpretq_u32_u8(uint8x16_t __v)
{
	return (uint32x4_t)__v;
}

_INTRINSATTR
static __inline uint64x2_t
vreinterpretq_u64_u8(uint8x16_t __v)
{
	return (uint64x2_t)__v;
}

_INTRINSATTR
static __inline uint8x16_t
vreinterpretq_u8_s32(int32x4_t __v)
{
	return (uint8x16_t)__v;
}

_INTRINSATTR
static __inline uint8x16_t
vreinterpretq_u8_u32(uint32x4_t __v)
{
	return (uint8x16_t)__v;
}

_INTRINSATTR
static __inline uint8x16_t
vreinterpretq_u8_u64(uint64x2_t __v)
{
	return (uint8x16_t)__v;
}

_INTRINSATTR
static __inline uint32x4_t
vsetq_lane_u32(uint32_t __x, uint32x4_t __v, uint8_t __i)
{
	__v[__neon_lane_index(__v, __i)] = __x;
	return __v;
}

_INTRINSATTR
static __inline uint64x2_t
vsetq_lane_u64(uint64_t __x, uint64x2_t __v, uint8_t __i)
{
	__v[__neon_lane_index(__v, __i)] = __x;
	return __v;
}

_INTRINSATTR
static __inline uint8x16_t
vshrq_n_u8(uint8x16_t __v, uint8_t __bits)
{
#ifdef __aarch64__
	return (uint8x16_t)__builtin_aarch64_lshrv16qi((int8x16_t)__v, __bits);
#else
	return (uint8x16_t)__builtin_neon_vshru_nv16qi((int8x16_t)__v, __bits);
#endif
}

_INTRINSATTR
static __inline int32x4_t
vsliq_n_s32(int32x4_t __vins, int32x4_t __vsh, uint8_t __bits)
{
#ifdef __aarch64__
	return (int32x4_t)__builtin_aarch64_ssli_nv4si(__vins, __vsh, __bits);
#else
	return (int32x4_t)__builtin_neon_vsli_nv4si(__vins, __vsh, __bits);
#endif
}

_INTRINSATTR
static __inline void
vst1q_u8(uint8_t *__p8, uint8x16_t __v)
{
#ifdef __aarch64__
	__builtin_aarch64_simd_qi *__p = (__builtin_aarch64_simd_qi *)__p8;

	__builtin_aarch64_st1v16qi(__p, (int8x16_t)__v);
#else
	__builtin_neon_qi *__p = (__builtin_neon_qi *)__p8;

	__builtin_neon_vst1v16qi(__p, (int8x16_t)__v);
#endif
}

#endif	/* _SYS_CRYPTO_AES_ARCH_ARM_ARM_NEON_H */
